
#include <inttypes.h>
#include <math.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "ks0108.h"
#include "corsiva_12.h"
#include "arial.h"
#include "usart.h"

#define MIN_X			0
#define MAX_X			127
#define MIN_Y			0
#define MAX_Y			63
#define MAX_SCORE		20
#define MIN_SPEED		1
#define MAX_SPEED		8
// paddle sizes */ 
#define P_X				1
#define P_Y				6
// ball size
#define BALL			1		
#define P_SPEED			4
#define BALL_SPEED		6
#define DELAY			200

/* two player scores */
int8_t score[2];

/* x axys and y axys ball velocities */
int8_t vx[3], vy[3], moved[3] = {0};

/* player paddles */
uint8_t posx[3], posy[3];

uint8_t text_height = 0, min_y=MIN_Y, intr_count=0;

/* the index of the player that lost */
int8_t gameOver;	

void initData(void) {
	min_y = MIN_Y + text_height;
	score[0] = score[1] = MAX_SCORE;
	posx[0] = MIN_X + 2*P_X;
	posx[1] = MAX_X - 2*P_X;
	posy[0] = posy[1] = (min_y + MAX_Y)/2;
	
	posx[2] = (MIN_X+MAX_X)/2;
	posy[2] = (min_y+MAX_Y)/2;
	vx[0] = vx[1] = 0;
	vy[0] = vy[1] = 0; 
	vy[2] = 0;
	vx[2] = BALL_SPEED;
	gameOver = -1;
}

void putScore(int8_t i) {
	int8_t sco = score[i];
	
	ks0108PutChar(sco/10+'0');
	sco = sco%10;
	
	ks0108PutChar(sco+'0');
	ks0108PutChar(' ');
}

void putScores(void) {
	/*
	ks0108GotoXY(MIN_X, MIN_Y);
	int8_t sco = intr_count;
	ks0108PutChar(sco/10+'0');
	sco = sco%10;
	
	ks0108PutChar(sco+'0');
	*/
	
	// write player 0 score
	ks0108GotoXY((MIN_X+MAX_X)/4, MIN_Y);
	putScore(0);
	// write player 1 score
	ks0108GotoXY(3*(MIN_X+MAX_X)/4, MIN_Y);
	putScore(1);
	
	ks0108DrawLine(MIN_X, min_y, MAX_X, min_y, BLACK);
}

void putField(uint8_t color) {
	ks0108DrawLine(MIN_X, min_y, MAX_X, min_y, color);
	//ks0108DrawLine(MAX_X, min_y, MAX_X, MAX_Y, color);
	ks0108DrawLine(MAX_X, MAX_Y, MIN_X, MAX_Y, color);
	//ks0108DrawLine(MIN_X, MAX_Y, MIN_X, min_y, color);
}

void putPaddle(uint8_t i, uint8_t color) {

	/* draw translated paddles for players */
	ks0108DrawRect(posx[i]-P_X, posy[i]-P_Y, 2*P_X, 2*P_Y, color);
}

void putBall(uint8_t color) {
	ks0108DrawCircle(posx[2], posy[2], 2*BALL, color);
} 

int8_t testBallOutside(int xmin, int xmax) {
	if (xmin <= MIN_X || xmax >= MAX_X)
		return 1;
	else 
		return 0;
}


int8_t testPingPong(int8_t p, int8_t xmin, int8_t ymin, 
					int8_t xmax, int8_t ymax) {
	
	int8_t pxmin, pxmax, pymin, pymax;
	pxmin = posx[p] - P_X-1;
	pxmax = posx[p] + P_X+1;
	pymin = posy[p] - P_Y-1;
	pymax = posy[p] + P_Y+1;
	
	// no y axys collision
	if ((ymax < pymin || pymax < ymin) || 
						 (xmax < pxmin || pxmax < xmin)) {
		return 0;
	}
	return 1;
}


void updateMoves(void) {
	int i;
	for (i=0; i<3; i++) {
		if (vx[i] || vy[i]) {
			moved[i] = 1;
			if (i<2) {
				putPaddle(i, WHITE);
			} else {
				putBall(WHITE);
			}
		}
		posx[i] += vx[i];
		posy[i] += vy[i];
	}
	
	vy[0] = vy[1] = 0;
}

void collisionLogic(void) {
	int8_t i;
	

	//  collision of paddles with horizontal walls
	for (i=0; i < 2; i++ ) {		
		if (posy[i]-P_Y <= min_y+1) { 
			posy[i] = min_y+P_Y+2;
		}
		else if (posy[i]+P_Y >= MAX_Y-1) {
			posy[i] = MAX_Y-P_Y-2;
		}
	}
	
	// collision of ball with horizontal walls
	if (posy[2]-BALL <= min_y+1) {
		posy[2] = min_y+BALL+2;
		vy[2] = -vy[2];
	}
	else if (posy[2]+BALL >= MAX_Y-1) {
		posy[2] = MAX_Y-BALL-2;
		vy[2] = -vy[2];
	}
	
	// colision of ball with paddles
	for (i=0; i < 2; i++ ) {		
		if (testPingPong(i, posx[2]-BALL, posy[2]-BALL, \
							posx[2]+BALL, posy[2]+BALL) > 0) {
			
			// move away ball, not to intersect paddle
			if (i==0)
				posx[2] = posx[i] + (P_X + BALL + 2) ;
				
			else 
				posx[2] = posx[i] - (P_X + BALL + 2) ;
			vx[2] = -vx[2];
			vy[2] = BALL_SPEED*(posy[2]-posy[i])/P_Y;
			break;
		}
	}
	
	/* if balls didn't collide with any of the paddles, 
	 * test exit of ball from field 
	 */
	if (i==2) {
		if (testBallOutside(posx[2]-BALL, posx[2]+BALL) > 0) {
			vy[2] = 0;
			posx[2] = (MIN_X+MAX_X)/2;
			posy[2] = (min_y+MAX_Y)/2;
			
			if (vx[2] < 0) {
				vx[2] = BALL_SPEED;
				score[0] --;
				if (score[0]==0)
					gameOver=0;
			} else {
				vx[2] = -BALL_SPEED;
				score[1] --;
				if (score[1]==0)
					gameOver=1;
			}
		} 
	}
}

void putMovedObjects(void) {
	uint8_t i;
	for(i=0; i< 3; i++) {
		if (moved[i]) {
			moved[i] = 0;
			if (i<2) {
				putPaddle(i, BLACK);
			} else {
				putBall(BLACK);
			}
		}
	}

}

void gameLogic(void) {
	/*erase old Score*/
	ks0108FillRect(0, 0, MIN_X, min_y, WHITE);
	/* clear field pixels */
		
	/* move ball and paddles */
	updateMoves();
	
	/* detect collisions between walls, ball, paddles */
	collisionLogic();
		
				
	/* show new score*/
	ks0108SelectFont(new_Font, ks0108ReadFontData, BLACK);
	putScores();
	
	putMovedObjects();

}


void endGame(void) {
	/* delete game field */
	ks0108ClearScreen();
	
	/* print end game message for the winner */
	ks0108SelectFont(Corsiva_12, ks0108ReadFontData, BLACK);
	ks0108GotoXY(10, (MIN_Y+MAX_Y)/2);
	ks0108Puts("Player ");
	ks0108PutChar(1-gameOver+'1');
	ks0108Puts(" wins");
	
	/* print nice round rectangle around text */
	text_height = ks0108GetFontHeight();
	ks0108DrawRoundRect(5, (MIN_Y+MAX_Y)/2-10, 117, 20, 8, BLACK);
}

int main(void) {	

	// Wait a little while the display starts up
	for(volatile uint16_t i=0; i<15000; i++);
	
	// Initialize the LCD
	ks0108Init(0);
	
	// Initialize the serial interface
	init_USART(0);
	usart_send(ACK);
	
	ks0108SelectFont(new_Font, ks0108ReadFontData, BLACK);	
	text_height = ks0108GetFontHeight();

	initData();
	putScores();
	putField(BLACK);
	
	putPaddle(0,BLACK); 
	putPaddle(1,BLACK);
	putBall(BLACK);
	
	/* start interrupt handling system */
	//sei();
	while(1) {	
		
		
		if (bit_is_set(UCSRA, RXC)) {
			uint8_t data = usart_recv();
			
			_delay_ms(DELAY);
			usart_send(ACK);
			
			uint8_t dir1, dir2;
			intr_count ++;
			unpack_move (data, &dir1, &dir2);
	
			if (dir1==UP) {
				vy[0] = P_SPEED;
			}
			
			if (dir1==DOWN) {
				vy[0] = -P_SPEED;
			}
			
			if (dir2==UP) {
				vy[1] = P_SPEED;
			}
			
			if (dir2==DOWN) {
				vy[1] = -P_SPEED;
			}
			
		} else {
			_delay_ms(DELAY);
		}
		
		gameLogic();		
	
		if(gameOver!=-1) {
					
			_delay_ms(2*DELAY);
			endGame();
			break;
		}
	}
	/*stop interrupt handling system */
	//cli();	
	while(1);
	
	return 0;
}

/*
ISR(USART_RXC_vect)
{
	_delay_ms(2*DELAY);
	usart_send(ACK);
	
	if(gameOver!=-1) {
		return;
	}

	intr_count ++;
	uint8_t data, dir1, dir2;
	
	// Fetch the received byte value 
	data = UDR;
	unpack_move (data, &dir1, &dir2);
	
	if (dir1==UP) {
		vy[0] = P_SPEED;
	}
	
	if (dir1==DOWN) {
		vy[0] = -P_SPEED;
	}
	
	if (dir2==UP) {
		vy[1] = P_SPEED;
	}
	
	if (dir2==DOWN) {
		vy[1] = -P_SPEED;
	}
	
	gameLogic();	
	if(gameOver!=-1) {
				
		_delay_ms(DELAY*2);
		endGame();
	}
	
} 
*/